# coding: utf8
# Debian Contributors data mining on file systems
#
# Copyright (C) 2013  Enrico Zini <enrico@debian.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from ..submission import Identifier
import os
import os.path
import pwd
import datetime
import logging

log = logging.getLogger(__name__)

__all__ = ["Filesystem"]

class Filesystem(object):
    """
    Collect and aggregate contribution data from file inode information, and
    build contribution informations out of it
    """
    def __init__(self):
        self.contribs = {}

    def scan_file(self, pathname):
        """
        Add an information from the inode information of a file
        """
        self.scan_stat(os.stat(pathname))

    def scan_stat(self, st):
        """
        Add an information from a stat structure
        """
        old = self.contribs.get(st.st_uid, None)
        if old is None:
            self.contribs[st.st_uid] = (st.st_mtime, st.st_mtime)
        else:
            self.contribs[st.st_uid] = (min(old[0], st.st_mtime), max(old[1], st.st_mtime))

    def scan_git_repo(self, gitdir):
        """
        Add information from refs files in the given git repo

        gitdir: pathname to the bare repository or the .git directory
        """
        scanroot = os.path.join(gitdir, "refs")
        log.debug("Starting git scanning at %s", scanroot)
        for root, dirs, files in os.walk(scanroot):
            for f in files:
                self.scan_file(os.path.join(root, f))

    def scan_svn_repo(self, svnroot):
        """
        Add information from commits in the given svn repo

        svnroot: pathname to the svn repository root dir
        """
        root = os.path.join(svnroot, "db/revs")
        for f in os.listdir(root):
            self.scan_file(os.path.join(root, f))

    def add_to_submission(self, submission, type, make_url=None):
        """
        Send all data we collected and aggregated to a Submission object

        submission: the Submission object that will get the data
        type: the contribution type
        make_url: if present, a function that will turn a pwd object into a URL
                  pointing to a list of contributions of this type for the user
        """
        for uid, stats in self.contribs.iteritems():
            try:
                pw = pwd.getpwuid(uid)
                ident = Identifier("login", pw.pw_name)
                begin = datetime.date.fromtimestamp(stats[0])
                end = datetime.date.fromtimestamp(stats[1])
                if make_url is not None:
                    url = make_url(pw)
                else:
                    url = None
                submission.add_contribution_data(ident, type, begin, end, url=url)
            except KeyError:
                pass
