# coding: utf8
# Debian Contributors data mining on emails
#
# Copyright (C) 2013  Enrico Zini <enrico@debian.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals
from ..submission import Identifier
from .stats import ContributorFrequencyCheck
from collections import namedtuple, defaultdict
import mailbox
import tempfile
import gzip
import shutil
import os.path
import email.utils
import datetime
import time

__all__ = ["MailboxPool"]

class CompressedMbox(mailbox.mbox):
    """
    Read-only access of a compressed mbox using a temporary file for the
    uncompressed version
    """
    def __init__(self, pathname):
        self.tempfile = tempfile.NamedTemporaryFile()
        with gzip.open(pathname) as fd:
            shutil.copyfileobj(fd, self.tempfile)
            self.tempfile.flush()
        mailbox.mbox.__init__(self, self.tempfile.name)

    def close(self):
        super(CompressedMbox, self).close()
        self.tmpfile.close()

def _collate_idents(gen, results):
    """
    Given a generator of (ident, ts), adding them to a
    dict { ident: (mints, maxts) }
    """
    for ident, ts in gen:
        old_range = results.get(ident, None)
        if old_range is None:
            results[ident] = ts, ts
        else:
            results[ident] = min(old_range[0], ts), max(old_range[1], ts)

def _get_mbox(pathname):
    """
    Create the right Mailbox object for a pathname
    """
    if os.path.isdir(pathname):
        return mailbox.Maildir(pathname)
    elif pathname.endswith(".gz"):
        return CompressedMbox(pathname)
    else:
        return mailbox.mbox(pathname)

ThreadInfo = namedtuple("ThreadInfo", ("email", "msgid", "date", "refs"))

class FolderScanner(object):
    def __init__(self, pathname):
        self.pathname = pathname
        self.folder = _get_mbox(pathname)

    def scan_known(self, address_map=None):
        """
        Scan the mailbox using a list of known addresses

        address_map: { email: Identifier } mapping

        Return a generator of (ident, ts).
        """
        for msg in self.folder:
            addr = msg.get("From", None)
            if addr is None: continue
            realname, emailaddr = email.utils.parseaddr(addr)
            if address_map:
                ident = address_map.get(emailaddr, None)
                if ident is None: continue
            else:
                ident = Identifier("email", emailaddr, realname)
            date = msg.get("Date", None)
            if date is None: continue
            date = email.utils.parsedate(date)
            if date is None: continue
            ts = time.mktime(date)
            yield ident, ts

    def scan_threads(self):
        """
        Scan the mailbox, generating a sequence of simplified threading info

        Each element generated is
        {
          msgid: Message ID
          date: date field of the message
          refs: Message ID of the first message this replies to (optional)
        }
        """
        for msg in self.folder:
            email = msg.get("From", None)
            if email is None: continue
            # Get message id
            msgid = msg.get("Message-ID", None)
            if msgid is None: continue
            # Get date
            date = msg.get("Date", None)
            if date is None: continue
            # Get references
            refs = msg.get("References", None)
            if refs:
                refs = refs.split()[0]
            yield ThreadInfo(email, msgid, date, refs)

class MailboxPool(object):
    """
    Collect and aggregate contribution data from one or more mail folders, in
    mbox or maildir formats
    """
    def __init__(self, pathnames=()):
        self.pathnames = list(pathnames)
        self.contribs = {}

    def add_pathname(self, pathname):
        """
        Add a mailbox pathname to the pool
        """
        self.pathnames.append(pathname)

    def scan_known_emails(self, address_map=None):
        """
        Consider any mail post from an email address whitelisted in address_map
        a contribution, and collect contribution data from this mailbox.

        address_map: { email: Identifier } mapping. If missing, all posts are
                     considered contributions
        """
        for pathname in self.pathnames:
            scanner = FolderScanner(pathname)
            _collate_idents(scanner.scan_known(address_map), self.contribs)

    def _gen_replies(self):
        """
        TODO: being worked on
        """
        # Collect thread info from all mailboxes, indexed by message ID
        by_msgid = {}
        for pathname in self.pathnames:
            scanner = FolderScanner(pathname)
            for tinfo in scanner.scan_threads():
                by_msgid[tinfo.msgid] = tinfo

        # Look for messages that replied to other messages
        for tinfo in by_msgid.itervalues():
            # Not a reply
            if tinfo.refs is None: continue
            # Get the message we reply to
            prev = by_msgid.get(tinfo.refs, None)
            if prev is None: continue
            realname, emailaddr = email.utils.parseaddr(tinfo.email)
            ident = Identifier("email", emailaddr, realname)
            date = email.utils.parsedate(prev.date)
            if date is None: continue
            ts = time.mktime(date)
            yield ident, ts

    def scan_replies(self, decay_days=7):
        """
        Look at replies to messages in a list and check contributor status
        using statistics on sustained contributions over a period of time
        """
        class ContributorDict(defaultdict):
            def __missing__(self, key):
                res = ContributorFrequencyCheck(key, decay_days)
                self[key] = res
                return res
        contributors = ContributorDict()
        for ident, ts in self._gen_replies():
            contributors[ident].add_stamp(ts)
        for c in contributors.itervalues():
            if c.is_contributor():
                self.contribs[c.ident] = c.contrib_range
            #import sys
            #c.test_dump_heat_function()
            #if c.is_contributor():
            #    print(c.ident, "IS contributor", file=sys.stderr)
            #else:
            #    print(c.ident, "NOT a contributor", file=sys.stderr)

    def add_to_submission(self, submission, type, make_url=None):
        """
        Send all data we collected and aggregated to a Submission object

        submission: the Submission object that will get the data
        type: the contribution type
        make_url: if present, a function that will turn an identity object into
                  a URL pointing to a list of contributions of this type for
                  the user
        """
        for ident, stats in self.contribs.iteritems():
            try:
                begin = datetime.date.fromtimestamp(stats[0])
                end = datetime.date.fromtimestamp(stats[1])
                if make_url is not None:
                    url = make_url(ident)
                else:
                    url = None
                submission.add_contribution_data(ident, type, begin, end, url=url)
            except KeyError:
                pass
